@@ -243,10 +243,17 @@ private slots:
     void multipleResponses();
     void contextObjectInOtherThreadWarning();
     void keepAliveTimeout();
+    void maxUrlSizeAllowed();
+    void maxTotalHeaderSizeAllowed();
+    void maxHeaderFieldSizeAllowed();
+    void maxNumberOfHeaderFieldsAllowed();
+    void maxBodySizeAllowed();
     void writeSequentialDevice();
     void writeMuchToSequentialDevice();
     void writeFromEmptySequentialDevice();
     void concurrentRequestsToSequentialDevice();
+    void timeoutConnection();
+    void useCanceledResponders();
 
 #if QT_CONFIG(localserver)
     void localSocket();
@@ -262,7 +269,9 @@ private:
     std::array<QNetworkAccessManager, 10> networkAccessManagers;
     QNetworkAccessManager &networkAccessManager = networkAccessManagers[0];
     ReplyObject replyObject;
+    std::vector<std::unique_ptr<QHttpServerResponder>> responders;
     std::optional<QSemaphore> readySem, routeSem;
+    std::optional<bool> canceledByClient;
 };
 
 struct CustomArg {
@@ -678,6 +687,28 @@ httpserver.route("/concurrent-request-back2", this,
                             }
                         });
             });
+
+    httpserver.route("/timeout-connection", this, [this](QHttpServerResponder &&responder) {
+        return QtConcurrent::run(
+                [this, r = std::make_shared<QHttpServerResponder>(std::move(responder))] {
+                    r->writeBeginChunked("text/plain"_ba);
+                    for (int i = 1; i < 200; ++i) {
+                        if (r->isResponseCanceled()) {
+                            canceledByClient.emplace(true);
+                            return;
+                        }
+                        QThread::sleep(10ms);
+                    }
+                    canceledByClient.emplace(false);
+                    r->writeEndChunked("chunk");
+                });
+    });
+
+    httpserver.route("/add-responder", this, [this](QHttpServerResponder &&responder) {
+        responder.writeBeginChunked("text/plain"_ba);
+        responder.writeChunk("chunk");
+        responders.emplace_back(std::make_unique<QHttpServerResponder>(std::move(responder)));
+    });
 #endif
 
 #if QT_CONFIG(localserver)
@@ -852,11 +883,14 @@ httpserver.route("/concurrent-request-back2", this,
 
 void tst_QHttpServer::init()
 {
+    canceledByClient.reset();
+    responders.clear();
 #if QT_CONFIG(ssl)
     QFETCH_GLOBAL(const bool, useSsl);
     if (useSsl && QTestPrivate::isSecureTransportBlockingTest())
         QSKIP("SslServer is blocking the test execution while trying to access the login keychain");
 #endif // QT_CONFIG(ssl)
+    networkAccessManager.clearConnectionCache();
 }
 
 void tst_QHttpServer::routeGet_data()
@@ -1089,7 +1123,7 @@ void tst_QHttpServer::routeGet()
     QFETCH(QString, type);
     QFETCH(QString, body);
     QString urlBase = useSsl ? sslUrlBase : clearUrlBase;
-    QNetworkRequest request = QNetworkRequest(urlBase.arg(url));
+    QNetworkRequest request = QNetworkRequest(QUrl{urlBase.arg(url)});
     request.setAttribute(QNetworkRequest::Http2AllowedAttribute, useHttp2);
 
     std::unique_ptr<QNetworkReply> reply(networkAccessManager.get(request));
@@ -1125,7 +1159,7 @@ void tst_QHttpServer::routeKeepAlive()
             .arg(static_cast<int>(req.method()));
     });
 
-    QNetworkRequest request(urlBase.arg("/keep-alive"));
+    QNetworkRequest request(QUrl{urlBase.arg("/keep-alive")});
     request.setRawHeader(QByteArray("Connection"), QByteArray("keep-alive"));
     request.setAttribute(QNetworkRequest::Http2AllowedAttribute, useHttp2);
 
@@ -1135,7 +1169,7 @@ void tst_QHttpServer::routeKeepAlive()
     if (QTest::currentTestFailed())
         return;
 
-    request.setUrl(urlBase.arg("/keep-alive?po=98"));
+    request.setUrl(QUrl{urlBase.arg("/keep-alive?po=98")});
     request.setRawHeader("CustomHeader", "1");
     request.setHeader(QNetworkRequest::ContentTypeHeader, "text/html"_ba);
 
@@ -1145,7 +1179,7 @@ void tst_QHttpServer::routeKeepAlive()
     if (QTest::currentTestFailed())
         return;
 
-    request = QNetworkRequest(urlBase.arg("/keep-alive"));
+    request = QNetworkRequest(QUrl{urlBase.arg("/keep-alive")});
     request.setRawHeader(QByteArray("Connection"), QByteArray("keep-alive"));
     request.setHeader(QNetworkRequest::ContentTypeHeader, "text/html"_ba);
     request.setAttribute(QNetworkRequest::Http2AllowedAttribute, useHttp2);
@@ -1271,7 +1305,7 @@ void tst_QHttpServer::routePost()
     QFETCH(QString, data);
     QFETCH(QString, body);
     QString urlBase = useSsl ? sslUrlBase : clearUrlBase;
-    QNetworkRequest request(urlBase.arg(url));
+    QNetworkRequest request(QUrl{urlBase.arg(url)});
     if (data.size())
         request.setHeader(QNetworkRequest::ContentTypeHeader, "text/html"_ba);
     request.setAttribute(QNetworkRequest::Http2AllowedAttribute, useHttp2);
@@ -1318,7 +1352,7 @@ void tst_QHttpServer::routeDelete()
     QFETCH(int, code);
     QFETCH(QString, type);
     QString urlBase = useSsl ? sslUrlBase : clearUrlBase;
-    QNetworkRequest request(urlBase.arg(url));
+    QNetworkRequest request(QUrl{urlBase.arg(url)});
     request.setAttribute(QNetworkRequest::Http2AllowedAttribute, useHttp2);
 
     std::unique_ptr<QNetworkReply> reply(networkAccessManager.deleteResource(request));
@@ -1342,7 +1376,7 @@ void tst_QHttpServer::routeExtraHeaders()
     QFETCH_GLOBAL(bool, useSsl);
     QFETCH_GLOBAL(bool, useHttp2);
     QString urlBase = useSsl ? sslUrlBase : clearUrlBase;
-    QNetworkRequest request(urlBase.arg("/extra-headers"));
+    QNetworkRequest request(QUrl{urlBase.arg("/extra-headers")});
     request.setAttribute(QNetworkRequest::Http2AllowedAttribute, useHttp2);
 
     std::unique_ptr<QNetworkReply> reply(networkAccessManager.get(request));
@@ -1367,7 +1401,7 @@ void tst_QHttpServer::getLongChunks()
     QFETCH_GLOBAL(bool, useSsl);
     QFETCH_GLOBAL(bool, useHttp2);
     QString urlBase = useSsl ? sslUrlBase : clearUrlBase;
-    QNetworkRequest request(urlBase.arg("/longChunks/"));
+    QNetworkRequest request(QUrl{urlBase.arg("/longChunks/")});
     request.setAttribute(QNetworkRequest::Http2AllowedAttribute, useHttp2);
 
     std::unique_ptr<QNetworkReply> reply(networkAccessManager.get(request));
@@ -1455,7 +1489,7 @@ void tst_QHttpServer::afterRequest()
     QFETCH_GLOBAL(bool, useSsl);
     QFETCH_GLOBAL(bool, useHttp2);
     QString urlBase = useSsl ? sslUrlBase : clearUrlBase;
-    QNetworkRequest request(urlBase.arg("/test-after-request"));
+    QNetworkRequest request(QUrl{urlBase.arg("/test-after-request")});
     request.setAttribute(QNetworkRequest::Http2AllowedAttribute, useHttp2);
 
     httpserver.addAfterRequestHandler(this, [] (const QHttpServerRequest &request,
@@ -1522,7 +1556,7 @@ void tst_QHttpServer::disconnectedInEventLoop()
     QFETCH_GLOBAL(bool, useSsl);
     QFETCH_GLOBAL(bool, useHttp2);
     QString urlBase = useSsl ? sslUrlBase : clearUrlBase;
-    QNetworkRequest request(urlBase.arg("/event-loop/"));
+    QNetworkRequest request(QUrl{urlBase.arg("/event-loop/")});
     request.setAttribute(QNetworkRequest::Http2AllowedAttribute, useHttp2);
 
     httpserver.route("/event-loop/", this, [] () {
@@ -1546,7 +1580,7 @@ void tst_QHttpServer::multipleRequests()
     QFETCH_GLOBAL(bool, useSsl);
     QFETCH_GLOBAL(bool, useHttp2);
     QString urlBase = useSsl ? sslUrlBase : clearUrlBase;
-    QNetworkRequest request(urlBase.arg("/do-not-move"));
+    QNetworkRequest request(QUrl{urlBase.arg("/do-not-move")});
     request.setAttribute(QNetworkRequest::Http2AllowedAttribute, useHttp2);
 
     httpserver.route("/do-not-move", this, [v = std::vector<int>{1, 2, 3}] () {
@@ -1959,6 +1993,204 @@ void tst_QHttpServer::keepAliveTimeout()
 #endif // QT_CONFIG(concurrent)
 }
 
+void tst_QHttpServer::maxUrlSizeAllowed()
+{
+#if QT_CONFIG(concurrent)
+    QFETCH_GLOBAL(bool, useSsl);
+    QFETCH_GLOBAL(bool, useHttp2);
+
+    QString urlBase = useSsl ? sslUrlBase : clearUrlBase;
+
+    QHttpServerConfiguration config;
+    config.setMaxUrlSize(30);
+    httpserver.setConfiguration(config);
+
+    auto cleanup = qScopeGuard([this] {
+        QHttpServerConfiguration config;
+        httpserver.setConfiguration(config);
+    });
+
+    QNetworkRequest request(QUrl(urlBase.arg(u"/req-and-resp-from-object"_s)));
+    request.setAttribute(QNetworkRequest::HttpPipeliningAllowedAttribute, true);
+    request.setAttribute(QNetworkRequest::Http2AllowedAttribute, useHttp2);
+    request.setHeader(QNetworkRequest::ContentTypeHeader, "text/plain");
+    std::unique_ptr<QNetworkReply> reply(networkAccessManager.post(request, "Hey"));
+
+    QTRY_VERIFY(reply->isFinished());
+    checkReply(reply.release(), "Hey");
+
+    config.setMaxUrlSize(3);
+    httpserver.setConfiguration(config);
+
+    reply.reset(networkAccessManager.post(request, "Hey"));
+    QTRY_VERIFY(reply->isFinished());
+    QCOMPARE(reply->attribute(QNetworkRequest::HttpStatusCodeAttribute).toInt(), 414);
+    QTest::qSleep(2000);
+
+#else
+    QSKIP("QtConcurrent is not available, skipping test");
+#endif // QT_CONFIG(concurrent)
+}
+
+void tst_QHttpServer::maxTotalHeaderSizeAllowed()
+{
+#if QT_CONFIG(concurrent)
+    QFETCH_GLOBAL(bool, useSsl);
+    QFETCH_GLOBAL(bool, useHttp2);
+
+    QString urlBase = useSsl ? sslUrlBase : clearUrlBase;
+
+    QHttpServerConfiguration config;
+    config.setMaxTotalHeaderSize(500);
+    httpserver.setConfiguration(config);
+
+    auto cleanup = qScopeGuard([this] {
+        QHttpServerConfiguration config;
+        httpserver.setConfiguration(config);
+    });
+
+    QNetworkRequest request(QUrl(urlBase.arg(u"/req-and-resp-from-object"_s)));
+    request.setAttribute(QNetworkRequest::HttpPipeliningAllowedAttribute, true);
+    request.setAttribute(QNetworkRequest::Http2AllowedAttribute, useHttp2);
+    request.setHeader(QNetworkRequest::ContentTypeHeader, "text/plain");
+    std::unique_ptr<QNetworkReply> reply(networkAccessManager.post(request, "Hey"));
+
+    QTRY_VERIFY(reply->isFinished());
+    checkReply(reply.release(), "Hey");
+
+    config.setMaxTotalHeaderSize(10);
+    httpserver.setConfiguration(config);
+
+    reply.reset(networkAccessManager.post(request, "Hey"));
+    QTRY_VERIFY(reply->isFinished());
+    QCOMPARE(reply->attribute(QNetworkRequest::HttpStatusCodeAttribute).toInt(), 431);
+    QTest::qSleep(2000);
+
+#else
+    QSKIP("QtConcurrent is not available, skipping test");
+#endif // QT_CONFIG(concurrent)
+}
+
+void tst_QHttpServer::maxHeaderFieldSizeAllowed()
+{
+#if QT_CONFIG(concurrent)
+    QFETCH_GLOBAL(bool, useSsl);
+    QFETCH_GLOBAL(bool, useHttp2);
+
+    QString urlBase = useSsl ? sslUrlBase : clearUrlBase;
+
+    QHttpServerConfiguration config;
+    config.setMaxHeaderFieldSize(500);
+    httpserver.setConfiguration(config);
+
+    auto cleanup = qScopeGuard([this] {
+        QHttpServerConfiguration config;
+        httpserver.setConfiguration(config);
+    });
+
+    QNetworkRequest request(QUrl(urlBase.arg(u"/req-and-resp-from-object"_s)));
+    request.setAttribute(QNetworkRequest::HttpPipeliningAllowedAttribute, true);
+    request.setAttribute(QNetworkRequest::Http2AllowedAttribute, useHttp2);
+    request.setHeader(QNetworkRequest::ContentTypeHeader, "text/plain");
+    request.setRawHeader("X-Big-Header", QByteArray("GonnaBeLarge").repeated(10));
+    std::unique_ptr<QNetworkReply> reply(networkAccessManager.post(request, "Hey"));
+
+    QTRY_VERIFY(reply->isFinished());
+    checkReply(reply.release(), "Hey");
+
+    config.setMaxHeaderFieldSize(100);
+    httpserver.setConfiguration(config);
+
+    reply.reset(networkAccessManager.post(request, "Hey"));
+    QTRY_VERIFY(reply->isFinished());
+    QCOMPARE(reply->attribute(QNetworkRequest::HttpStatusCodeAttribute).toInt(), 431);
+    QTest::qSleep(2000);
+
+#else
+    QSKIP("QtConcurrent is not available, skipping test");
+#endif // QT_CONFIG(concurrent)
+}
+
+void tst_QHttpServer::maxNumberOfHeaderFieldsAllowed()
+{
+#if QT_CONFIG(concurrent)
+    QFETCH_GLOBAL(bool, useSsl);
+    QFETCH_GLOBAL(bool, useHttp2);
+
+    QString urlBase = useSsl ? sslUrlBase : clearUrlBase;
+
+    QHttpServerConfiguration config;
+    config.setMaxNumberOfHeaderFields(120);
+    httpserver.setConfiguration(config);
+
+    auto cleanup = qScopeGuard([this] {
+        QHttpServerConfiguration config;
+        httpserver.setConfiguration(config);
+    });
+
+    QNetworkRequest request(QUrl(urlBase.arg(u"/req-and-resp-from-object"_s)));
+    request.setAttribute(QNetworkRequest::HttpPipeliningAllowedAttribute, true);
+    request.setAttribute(QNetworkRequest::Http2AllowedAttribute, useHttp2);
+    request.setHeader(QNetworkRequest::ContentTypeHeader, "text/plain");
+    for (qsizetype i = 0; i < 100; ++i)
+        request.setRawHeader(QByteArray("X-Header-").append(QByteArray::number(i)), "Dummydata");
+    std::unique_ptr<QNetworkReply> reply(networkAccessManager.post(request, "Hey"));
+
+    QTRY_VERIFY(reply->isFinished());
+    checkReply(reply.release(), "Hey");
+
+    config.setMaxNumberOfHeaderFields(100);
+    httpserver.setConfiguration(config);
+
+    reply.reset(networkAccessManager.post(request, "Hey"));
+    QTRY_VERIFY(reply->isFinished());
+    QCOMPARE(reply->attribute(QNetworkRequest::HttpStatusCodeAttribute).toInt(), 431);
+    QTest::qSleep(2000);
+
+#else
+    QSKIP("QtConcurrent is not available, skipping test");
+#endif // QT_CONFIG(concurrent)
+}
+
+void tst_QHttpServer::maxBodySizeAllowed()
+{
+#if QT_CONFIG(concurrent)
+    QFETCH_GLOBAL(bool, useSsl);
+    QFETCH_GLOBAL(bool, useHttp2);
+
+    QString urlBase = useSsl ? sslUrlBase : clearUrlBase;
+
+    QHttpServerConfiguration config;
+    config.setMaxBodySize(500);
+    httpserver.setConfiguration(config);
+
+    auto cleanup = qScopeGuard([this] {
+        QHttpServerConfiguration config;
+        httpserver.setConfiguration(config);
+    });
+
+    QNetworkRequest request(QUrl(urlBase.arg(u"/req-and-resp-from-object"_s)));
+    request.setAttribute(QNetworkRequest::HttpPipeliningAllowedAttribute, true);
+    request.setAttribute(QNetworkRequest::Http2AllowedAttribute, useHttp2);
+    request.setHeader(QNetworkRequest::ContentTypeHeader, "text/plain");
+    std::unique_ptr<QNetworkReply> reply(networkAccessManager.post(request, "Here's the body"));
+
+    QTRY_VERIFY(reply->isFinished());
+    checkReply(reply.release(), "Here's the body");
+
+    config.setMaxBodySize(10);
+    httpserver.setConfiguration(config);
+
+    reply.reset(networkAccessManager.post(request, "Here's the body"));
+    QTRY_VERIFY(reply->isFinished());
+    QCOMPARE(reply->attribute(QNetworkRequest::HttpStatusCodeAttribute).toInt(), 413);
+    QTest::qSleep(2000);
+
+#else
+    QSKIP("QtConcurrent is not available, skipping test");
+#endif // QT_CONFIG(concurrent)
+}
+
 void tst_QHttpServer::writeSequentialDevice()
 {
     QFETCH_GLOBAL(bool, useSsl);
@@ -2047,6 +2279,54 @@ void tst_QHttpServer::concurrentRequestsToSequentialDevice()
 #endif // QT_CONFIG(concurrent)
 }
 
+void tst_QHttpServer::timeoutConnection()
+{
+    QFETCH_GLOBAL(bool, useSsl);
+    QFETCH_GLOBAL(bool, useHttp2);
+    QVERIFY(!canceledByClient.has_value());
+
+    QString urlBase = useSsl ? sslUrlBase : clearUrlBase;
+    const QUrl requestUrl(urlBase.arg("/timeout-connection"));
+    QNetworkRequest req(requestUrl);
+    req.setAttribute(QNetworkRequest::Http2AllowedAttribute, useHttp2);
+    req.setTransferTimeout(1s);
+    std::unique_ptr<QNetworkReply> reply(networkAccessManager.get(req));
+
+    QSignalSpy spy(reply.get(), &QNetworkReply::finished);
+    spy.wait(3s);
+    QTRY_VERIFY(canceledByClient.has_value());
+    QCOMPARE(spy.count(), 1);
+    QVERIFY(reply->isFinished());
+    QCOMPARE(reply->error(), QNetworkReply::TimeoutError);
+    QVERIFY(canceledByClient.value());
+}
+
+void tst_QHttpServer::useCanceledResponders()
+{
+    QFETCH_GLOBAL(bool, useSsl);
+    QFETCH_GLOBAL(bool, useHttp2);
+
+    QString urlBase = useSsl ? sslUrlBase : clearUrlBase;
+    const QUrl requestUrl(urlBase.arg("/add-responder"));
+    QNetworkRequest req(requestUrl);
+    req.setAttribute(QNetworkRequest::Http2AllowedAttribute, useHttp2);
+    req.setTransferTimeout(10ms);
+    std::unique_ptr<QNetworkReply> reply1(networkAccessManager.get(req));
+    std::unique_ptr<QNetworkReply> reply2(networkAccessManager.get(req));
+
+    QTRY_VERIFY(reply1->isFinished());
+    QTRY_VERIFY(reply2->isFinished());
+
+    for (auto &responder : responders)
+        responder->writeChunk("chunk");
+
+    for (auto &responder : responders)
+        responder->writeEndChunked("end");
+
+    for (auto &responder : responders)
+        QVERIFY(responder->isResponseCanceled());
+}
+
 QT_END_NAMESPACE
 
 Q_DECLARE_METATYPE(CustomArg);
